# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

__all__ = ['IcePAPBackup']

import configparser
import time
import logging
import os
from icepap import IcePAPController


KEYNOTFOUNDIN1 = 'KeyNotFoundInBackup'       # KeyNotFound for dictDiff
KEYNOTFOUNDIN2 = 'KeyNotFoundInIcePAP'       # KeyNotFound for dictDiff
UNKNOWN = 'Unknown'


def dict_cfg(first, second):
    """
    Return a dict of keys that differ with another config object.  If a value
    is not found in one fo the configs, it will be represented by KEYNOTFOUND.
    @param first:   Fist configuration to diff.
    @param second:  Second configuration to diff.
    @return diff:   Dict of Key => (first.val, second.val)
    """

    diff = {}
    sd1 = set(first)
    sd2 = set(second)
    # Keys missing in the second dict
    for key in sd1.difference(sd2):
        diff[key] = (first[key], KEYNOTFOUNDIN2)
    # Keys missing in the first dict
    for key in sd2.difference(sd1):
        diff[key] = (KEYNOTFOUNDIN1, second[key])
    # Check for differences
    for key in sd1.intersection(sd2):
        value1 = first[key].upper().strip()
        value2 = second[key].upper().strip()
        if value1 != value2:
            diff[key] = (first[key], second[key])
    return diff


class IcePAPBackup:
    """
    Class to create/restore IcePAP backups based on Ethernet communication.
    """

    def __init__(self, host='', port=5000, timeout=3, cfg_file=''):
        log_name = '{0}.IcePAPBackup'.format(__name__)
        self.log = logging.getLogger(log_name)
        self._cfg_ipap = configparser.ConfigParser()
        self._cfg_bkp = None
        if cfg_file != '':
            self._cfg_bkp = configparser.ConfigParser()
            self._cfg_bkp.read(cfg_file)
            if host == '':
                host = self._cfg_bkp.get('SYSTEM', 'HOST')
            port = int(self._cfg_bkp.get('SYSTEM', 'PORT'))
        self._host = host
        self._port = port
        self._ipap = IcePAPController(host, port, timeout, auto_axes=True)

    def _add_axis(self, axis):
        """
        Method to add axis backup.
        :param axis: int
        :return: None
        """
        section_name = 'AXIS_{0}'.format(axis)
        self._cfg_ipap.add_section(section_name)
        # Active
        active = self._ipap[axis].active
        self._cfg_ipap.set(section_name, 'ACTIVE', str(active))
        if not active:
            self.log.warning('Driver {0} is not active, '
                             'some commands will fail.'.format(axis))
        # Version
        ver = self._ipap[axis].ver
        keys = list(ver['SYSTEM']['DRIVER'].keys())
        keys.sort()
        for key in keys:
            option = 'VER_{0}'.format(key)
            value = str(ver['SYSTEM']['DRIVER'][key][0])
            self._cfg_ipap.set(section_name, option, value)
        ver_drv = ver.driver[0]
        if ver_drv < 3:
            self.log.info('The version {0} does not support all command.'
                          'The script will generate warning'
                          'messages'.format(ver.driver))
        # Configuration
        ipap_cfg = self._ipap[axis].get_cfg()
        keys = list(ipap_cfg.keys())
        keys.sort()
        for key in keys:
            option = 'CFG_{0}'.format(key)
            value = ipap_cfg[key]
            self._cfg_ipap.set(section_name, option, value)

        # Attributes can fail on reading, but we save it anyway with UNKNOWN
        # value
        attrs = ['velocity', 'name', 'acctime', 'pcloop', 'indexer', 'infoa',
                 'infob', 'infoc', 'pos', 'pos_measure', 'pos_shftenc',
                 'pos_tgtenc', 'pos_ctrlenc', 'pos_encin', 'pos_inpos',
                 'pos_absenc', 'pos_motor', 'pos_sync', 'enc', 'enc_measure',
                 'enc_shftenc', 'enc_tgtenc', 'enc_ctrlenc', 'enc_encin',
                 'enc_inpos', 'enc_absenc', 'enc_motor', 'enc_sync', 'id']

        # Attributes can fail on reading because they are on version 3.17
        # This attributes won on the backup file if the version is < 3
        v3_attrs = ['cswitch', 'velocity_min', 'velocity_max', 'ecam',
                    'outpos', 'outpaux', 'syncpos', 'syncaux']

        attrs += v3_attrs
        attrs.sort()

        for attr in attrs:
            if attr in v3_attrs and ver_drv < 3:
                continue
            try:
                value = self._ipap[axis].__getattribute__(attr)
            except Exception as e:
                self.log.error('Error on reading axis {0} {1}: '
                               '{2}'.format(axis, attr, repr(e)))
                value = UNKNOWN
            self._cfg_ipap.set(section_name, attr, repr(value))

        # External Disable. Valid for FW < 3
        if ver_drv < 3:
            try:
                value = eval(self._ipap[axis].send_cmd('?DISDIS')[0])
            except Exception as e:
                self.log.error('Error on reading axis {0} {1}: '
                               '{2}'.format(axis, 'DISDIS', repr(e)))
                value = UNKNOWN
            self._cfg_ipap.set(section_name, 'DISDIS', repr(value))

    def _save_to_file(self, filename):
        abspath = os.path.abspath(filename)
        self.log.info('Saving backup on: {0}'. format(abspath))
        with open(abspath, 'w') as f:
            f.write('# File auto-generated by IcePAPBackup class.\n\n')
            self._cfg_ipap.write(f)

    def _add_system(self):
        section_name = 'SYSTEM'
        self._cfg_ipap.add_section(section_name)
        self._cfg_ipap.set(section_name, 'HOST', self._host)
        self._cfg_ipap.set(section_name, 'PORT', str(self._port))
        ver = str(self._ipap.ver['SYSTEM']['VER'][0])
        self._cfg_ipap.set(section_name, 'VERSION', ver)

    def _add_general(self):
        section_name = 'GENERAL'
        self._cfg_ipap.add_section(section_name)
        self._cfg_ipap.set(section_name, 'DATE', time.strftime('%Y/%m/%d'))
        self._cfg_ipap.set(section_name, 'TIME', time.strftime('%H:%M:%S +%z'))

    def _add_controller(self):
        section_name = 'CONTROLLER'
        self._cfg_ipap.add_section(section_name)
        # Version
        ver = self._ipap.ver
        keys = list(ver['SYSTEM']['CONTROLLER'].keys())
        keys.sort()
        for key in keys:
            option = 'VER_{0}'.format(key)
            value = str(ver['SYSTEM']['CONTROLLER'][key][0])
            self._cfg_ipap.set(section_name, option, value)

    def do_backup(self, filename='', axes=[], save=True, general=True):
        axes.sort()
        if general:
            self._add_general()
        self._add_system()
        self._add_controller()
        if len(axes) == 0:
            axes = self._ipap.axes
            axes.sort()
        for axis in axes:
            self._add_axis(axis)
        if save:
            self._save_to_file(filename)

    def do_check(self, axes=[]):
        self._cfg_bkp.pop('GENERAL')
        sections = self._cfg_bkp.sections()
        sections.pop(sections.index('SYSTEM'))
        sections.pop(sections.index('CONTROLLER'))
        for axis in axes:
            section = 'AXIS_{0}'.format(axis)
            try:
                sections.pop(sections.index(section))
            except Exception:
                raise ValueError('There is not backup for the axis '
                                 '{0}'.format(axis))
        if len(axes) > 0:
            for section in sections:
                self._cfg_bkp.pop(section)
        else:
            for section in sections:
                axis = int(section.split('_')[1])
                axes.append(axis)
        self.log.info('Checking IcePAP {0} axes: {1}'.format(self._host,
                                                             repr(axes)))
        self.do_backup(axes=axes, save=False, general=False)
        total_diff = {}
        if self._cfg_bkp == self._cfg_ipap:
            self.log.info('No differences found')
        else:
            sections = self._cfg_bkp.sections()
            for section in sections:
                diff = dict_cfg(self._cfg_bkp[section], self._cfg_ipap[
                    section])
                if len(diff) > 0:
                    total_diff[section] = diff
            self.log.info('Differences found: {0}'.format(repr(total_diff)))
        return total_diff

    def do_autofix(self, diff):
        """
        Solve inconsistencies in IcePAP configuration registers.

        :param diff: Differences dictiomnary.
        :return:
        """
        self.active_axes(force=True)
        time.sleep(2)
        sections = list(diff.keys())
        axes = []
        for section in sections:
            if 'AXIS_' in section:
                axis = int(section.split('_')[1])
                axes.append(axis)
        axes.sort()
        for axis in axes:
            section = 'AXIS_{0}'.format(axis)
            registers = diff[section]
            for register in registers:
                if 'ver' in register:
                    continue
                if 'cfg' in register:
                    continue

                value_bkp, value_ipap = diff[section][register]

                if UNKNOWN in value_bkp:
                    continue

                # Check DISDIS configuration
                if register.lower() == 'disdis':
                    try:
                        if 'KeyNot' in value_bkp:
                            # Version backup > 3
                            value = diff[section]['cfg_extdisable'][0]
                            if value.lower() == 'none':
                                cmd = 'DISDIS 1'
                            else:
                                cmd = 'DISDIS 0'
                            self._ipap[axis].send_cmd(cmd)
                            self.log.info('Fixed axis {0} disdis '
                                          'configuration: cfg_extdisable({1})'
                                          ' -> {2}'.format(axis, value, cmd))
                        else:
                            # Version backup < 3:
                            value_bkp = eval(value_bkp)
                            self._ipap[axis].send_cmd('config')
                            value = ['Disable', 'NONE'][value_bkp]
                            cfg = 'EXTDISABLE {0}'.format(value)
                            self._ipap[axis].set_cfg(cfg)
                            cmd_str = 'config conf{0:03d}'.format(axis)
                            self._ipap[axis].send_cmd(cmd_str)

                            self.log.info('Fixed axis {0} disdis '
                                          'configuration: DISDIS {1} -> '
                                          'cfg_extdisable {2}'
                                          .format(axis, value_bkp, value))
                    except Exception as e:
                        if self._ipap[axis].mode != 'oper':
                            self._ipap[axis].send_cmd('config')
                        self.log.error('Cannot fix axis {0} disdis '
                                       'configuration: bkp({1}) icepap({2}).'
                                       ' Error {3}'.format(axis, value_bkp,
                                                           value_ipap, e))

                if 'KeyNot' in value_ipap or 'KeyNot' in value_bkp:
                    continue

                if register.startswith('enc') and register != 'enc_encin':
                    self.log.warning('Skip axis {0} {1}: bkp({2}) '
                                     'icepap({3})'.format(axis, register,
                                                          value_bkp,
                                                          value_ipap))
                    continue
                if register.startswith('pos'):
                    self.log.warning('Skip axis {0} {1}: bkp({2}) '
                                     'icepap({3})'.format(axis, register,
                                                          value_bkp,
                                                          value_ipap))
                    continue

                try:
                    value = eval(value_bkp)
                    if register == 'velocity':
                        acctime = self._ipap[axis].acctime
                        self._ipap[axis].velocity = value
                        self._ipap[axis].acctime = acctime
                    else:
                        self._ipap[axis].__setattr__(register, value)

                    self.log.info('Fixed axis {0} {1}: bkp({2}) '
                                  'icepap({3})'.format(axis, register,
                                                       value_bkp, value_ipap))
                except Exception as e:
                    self.log.error('Cannot fix axis {0} {1}: bkp({2}) '
                                   'icepap({3}). '
                                   'Error {4})'.format(axis, register,
                                                       value_bkp, value_ipap,
                                                       e))
        self.active_axes()

    def active_axes(self, axes=[], force=False):
        sections = self._cfg_bkp.sections()
        for section in sections:
            if section in ['GENERAL', 'SYSTEM', 'CONTROLLER']:
                continue
            axis = int(section.split('_')[1])
            if axes == [] or axis in axes:
                active = self._cfg_bkp.getboolean(section, 'ACTIVE')
                self._ipap[axis].send_cmd('config')
                if force:
                    cfg = 'ACTIVE YES'
                else:
                    cfg = 'ACTIVE {0}'.format(['NO', 'YES'][active])
                self.log.info('Axis {0}: {1}'.format(axis, cfg))
                self._ipap[axis].set_cfg(cfg)
                cmd = 'config conf{0:03d}'.format(axis)
                self._ipap[axis].send_cmd(cmd)



# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import weakref
import struct
import collections
from .vdatalib import vdata, ADDRUNSET, POSITION, PARAMETER, SLOPE, DWORD, \
    FLOAT
from .utils import State
from .fwversion import FirmwareVersion

__all__ = ['IcePAPAxis']


class IcePAPAxis:
    """
    The IcePAP axis class contains the common IcePAP ASCii API for any
    IcePAP axis. The methods here implemented correspond to those
    at the axis level.
    """
    def __init__(self, ctrl, axis_nr):
        ref = weakref.ref(ctrl)
        self._ctrl = ref()
        self._axis_nr = axis_nr

        # if self._axis_nr != self.addr:
        #     msg = 'Initialization error: axis_nr {0} != adr {1}'.format(
        #         self._axis_nr, self.addr)
        #     raise RuntimeError(msg)

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self._axis_nr)

    def __str__(self):
        return 'IcePAPAxis {} on {}'.format(self._axis_nr, self._ctrl)

    @staticmethod
    def get_ushort_list(ldata, dtype='FLOAT'):
        dtype = dtype.upper()
        if dtype == 'DWORD':
            pack_format = '{0}i'
            factor = 2
        elif dtype == 'FLOAT':
            pack_format = '{0}f'
            factor = 2
        elif dtype == 'DFLOAT':
            pack_format = '{0}d'
            factor = 4
        elif dtype == 'BYTE':
            pack_format = '{0}b'
            factor = 0.5
        else:
            raise ValueError('dtype is not valid')

        data_len = len(ldata)
        data_pck = struct.pack(pack_format.format(data_len), *ldata)

        unpack_format = '{0}H'.format(int(data_len * factor))
        lushorts = struct.unpack(unpack_format, data_pck)
        return lushorts

    @staticmethod
    def get_dump_values(raw_table, dtype='FLOAT'):
        # TODO: use the memory map.
        values = []
        for raw_value in raw_table:
            try:
                id_value, _, value = raw_value.strip().split(':')
            except Exception:
                raise RuntimeError('There are not values loaded on the '
                                   'ecam table.')
            values.append(float(value))
            last_id_data, len_data = id_value.split('/')
        last_id_data = int(last_id_data)
        len_data = int(len_data)
        return values, last_id_data, len_data

    def _get_dump_table(self, cmd, dtype='FLOAT'):
        MAX_SUBSET_SIZE = 200
        table = []
        start_pos = 0
        while True:
            raw_table = self.send_cmd(cmd.format(MAX_SUBSET_SIZE, start_pos))
            values, last_id, len_data = self.get_dump_values(raw_table,
                                                             dtype)
            table += values
            # The data dumped from list position table has a bug the total
            # len is not correct
            if 'LISTDAT' in cmd.upper():
                len_data -= 1
            if int(last_id) == int(len_data):
                break
            start_pos = last_id + 1
        return table

    @property
    def addr(self):
        """
        Get the axis number (IcePAP user manual pag. 49).

        :return: int
        """
        return int(self.send_cmd('?ADDR')[0])

    @property
    def active(self):
        """
        Get if the axis is active (IcePAP user manual pag. 47).

        :return: bool
        """
        ans = self.send_cmd('?ACTIVE')[0].upper()
        return ans == 'YES'

    @property
    def mode(self):
        """
        Return the current mode of the axis: CONFIG, OPER, PROG, TEST, FAIL
        (IcePAP user manual pag. 91).

        :return: str
        """
        return self.send_cmd('?MODE')[0]

    @property
    def status(self):
        """
        Return axis status word 32-bits (IcePAP user manual pag. 128).

        :return: int
        """
        return int(self.send_cmd('?STATUS')[0], 16)

    @property
    def state(self):
        """
        Read the axis status and return a util.State object.

        :return: State
        """
        return State(self.status)

    @property
    def state_present(self):
        """
        Check if the present flag is active.

        :return: bool
        """
        return self.state.is_present()

    @property
    def state_alive(self):
        """
        Check if the alive flag is active.

        :return: bool
        """
        return self.state.is_alive()

    @property
    def state_mode_code(self):
        """
        Return the current mode.

        :return: int
        """
        return self.state.get_mode_code()

    @property
    def state_mode_str(self):
        """
        Return the current mode.

        :return: str
        """
        return self.state.get_mode_str()

    @property
    def state_disabled(self):
        """
        Check if the disable flag is active.

        :return: bool
        """
        return self.state.is_disabled()

    @property
    def state_disable_code(self):
        """
        Return the disable code.

        :return: int
        """
        return self.state.get_disable_code()

    @property
    def state_disable_str(self):
        """
        Return the disable string.

        :return: str
        """
        return self.state.get_disable_str()

    @property
    def state_indexer_code(self):
        """
        Return the indexer code.

        :return: int
        """
        return self.state.get_indexer_code()

    @property
    def state_indexer_str(self):
        """
        Return the indexer string.

        :return: str
        """
        return self.state.get_indexer_str()

    @property
    def state_moving(self):
        """
        Check if the moving flag is active.

        :return: bool
        """
        return self.state.is_moving()

    @property
    def state_ready(self):
        """
        Check if the ready flag is active.

        :return: bool
        """
        return self.state.is_ready()

    @property
    def state_settling(self):
        """
        Check if the settling flag is active.

        :return: bool
        """
        return self.state.is_settling()

    @property
    def state_outofwin(self):
        """
        Check if the outofwin flag is active.

        :return: bool
        """
        return self.state.is_outofwin()

    @property
    def state_warning(self):
        """
        Check if the warning flag is active.

        :return: bool
        """
        return self.state.is_warning()

    @property
    def state_stop_code(self):
        """
        Return the stop code.

        :return: int
        """
        return self.state.get_stop_code()

    @property
    def state_stop_str(self):
        """
        Return the stop string.

        :return: str
        """
        return self.state.get_stop_str()

    @property
    def state_limit_positive(self):
        """
        Check if the flag limit_positive is active.

        :return: bool
        """
        return self.state.is_limit_positive()

    @property
    def state_limit_negative(self):
        """
        Check if the flag limit_negative is active.

        :return: bool
        """
        return self.state.is_limit_negative()

    @property
    def state_inhome(self):
        """
        Chekc if the home flag is active.

        :return: bool
        """
        return self.state.is_inhome()

    @property
    def state_5vpower(self):
        """
        Check if the auxiliary power is On.

        :return: bool
        """
        return self.state.is_5vpower()

    @property
    def state_verserr(self):
        """
        Check if the vererr flag is active.

        :return: bool
        """
        return self.state.is_verserr()

    @property
    def state_poweron(self):
        """
        Check if the flag poweron is active.

        :return: bool
        """
        return self.state.is_poweron()

    @property
    def state_info_code(self):
        """
        Return the info code.

        :return: int
        """
        return self.state.get_info_code()

    @property
    def vstatus(self):
        """
        Return the axis status as multi-line verbose answer (IcePAP user
        manual pag. 146).

        :return: str
        """
        return '\n'.join(self.send_cmd('?VSTATUS'))

    @property
    def stopcode(self):
        """
        Return axis stop code word 16-bits (IcePAP user manual pag. 130).

        :return: int
        """
        return int(self.send_cmd('?STOPCODE')[0], 16)

    @property
    def vstopcode(self):
        """
        Return the message corresponding to the last motion's stop code
        (IcePAP user manual pag. 147).

        :return: str
        """
        return ' '.join(self.send_cmd('?VSTOPCODE'))

    @property
    def alarm(self):
        """
        Get if the axis is in alarm condition (IcePAP user manual pag. 50).

        :return: [bool, str]
        """
        ans = self.send_cmd('?ALARM')
        if len(ans) == 1:
            result = (False, '')
        else:
            result = (True, ' '.join(ans))
        return result

    @property
    def warning(self):
        """
        Return a list of string describing warning conditions (IcePAP user
        manual pag. 150).

        :return: [str]
        """
        return self.send_cmd('?WARNING')

    @property
    def wtemp(self):
        """
        Get the set temperature warning value (IcePAP user manual pag. 151).

        :return: float
        """
        return float(self.send_cmd('?WTEMP')[0])

    @wtemp.setter
    def wtemp(self, value):
        """
        Set the temperature warning value (IcePAP user manual pag. 151).

        :param value: float
        """
        cmd = 'WTEMP {0}'.format(value)
        self.send_cmd(cmd)

    @property
    def config(self):
        """
        Get the identifier of the last valid configuration. To change the
        configuration use set_config method (IcePAP user manual pag. 59).

        :return: str
        """
        config = self.send_cmd('?CONFIG')
        if config is None:
            config = ""
        else:
            config = config[0]
        return config

    @property
    def cswitch(self):
        """
        Get the limit switches configuration (IcePAP user manual pag. 60).

        :return: str
        """
        return self.send_cmd('?CSWITCH')[0]

    @cswitch.setter
    def cswitch(self, value):
        """
        Set the limit switches configuration (IcePAP user manual pag. 60).

        :param value: str [Normal, Smart, Sticky]
        """
        cmd = 'CSWITCH {0}'.format(value)
        self.send_cmd(cmd)

    @property
    def ver(self):
        """
        Get the version of the all driver modules: Driver, DSP, FPGA, PCB, IO
        (IcePAP user manual pag. 144).

        :return: dict{module: (ver, date)}
        """
        ans = self.send_cmd('?VER INFO')
        return FirmwareVersion(ans, True)

    @property
    def name(self):
        """
        Get the axis name (Icepap user manual pag. 95).

        :return: str
        """
        value = self.send_cmd('?NAME')
        if isinstance(value, list):
            value = value[0]
        return value

    @name.setter
    def name(self, value):
        """
        Set the axis name. It must lower than 20 characters (Icepap user manual
        pag. 95).

        :param value: str
        """
        if len(value) >= 20:
            raise ValueError('Name too long, max size 20 characters')
        cmd = 'NAME {0}'.format(value)
        self.send_cmd(cmd)

    @property
    def id(self):
        """
        Get hardware ID and the serial number (Icepap user manual pag. 80).

        :return: (str HW ID, str SN)
        """
        hw_id = self.send_cmd('?ID HW')[0]
        sn_id = self.send_cmd('?ID SN')[0]
        return hw_id, sn_id

    @property
    def post(self):
        """
        Get the result of the power-on self test. Zero means there were not
        errors (IcePAP user manual pag. 110).

        :return: int
        """
        return int(self.send_cmd('?POST')[0])

    @property
    def power(self):
        """
        Get if the axis is ON (IcePAP user manual pag. 111).

        :return: bool
        """
        return self.send_cmd('?POWER')[0] == 'ON'

    @power.setter
    def power(self, value):
        """
        Set the power of the axis (IcePAP user manual pag. 111).

        :param value: bool
        """
        cmd = 'POWER {0}'.format(['OFF', 'ON'][value])
        self.send_cmd(cmd)

    @property
    def auxps(self):
        """
        Get if the auxiliary power supply is ON (IcePAP user manual pag. 52).

        :return: bool
        """
        return self.send_cmd('?AUXPS')[0] == 'ON'

    @auxps.setter
    def auxps(self, value):
        """
        Set the auxiliary power supply of the axis (Icepap user manual pag.
        52).

        :param value: bool
        """
        cmd = 'AUXPS {0}'.format(['OFF', 'ON'][value])
        self.send_cmd(cmd)

    @property
    def meas_vcc(self):
        """
        Measured value of the main power supply (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('VCC')

    @property
    def meas_vm(self):
        """
        Measured value of the motor voltage (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('VM')

    @property
    def meas_i(self):
        """
        Measured value of the motor current (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('I')

    @property
    def meas_ia(self):
        """
        Measured value of the phase a current (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('IA')

    @property
    def meas_ib(self):
        """
        Measured value of the phase b current (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('IB')

    @property
    def meas_ic(self):
        """
        Measured value of the phase c current (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('IC')

    @property
    def meas_r(self):
        """
        Measured value of the motor resistance (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('R')

    @property
    def meas_ra(self):
        """
        Measured value of the phase a resistance (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('RA')

    @property
    def meas_rb(self):
        """
        Measured value of the phase b resistance (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('RB')

    @property
    def meas_rc(self):
        """
        Measured value of the phase c resistance (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('RC')

    @property
    def meas_t(self):
        """
        Measured value of the board temperature (IcePAP user manual pag. 89).

        :return: float
        """
        return self.meas('T')

    @property
    def meas_rt(self):
        """
        Measured value of the power supply temperature (IcePAP user manual
        pag. 89).

        :return: float
        """
        return self.meas('RT')

    @property
    def pos(self):
        """
        Read the axis nominal position pointer (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('AXIS')

    @pos.setter
    def pos(self, value):
        """
        Set the axis nominal position value (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('AXIS', value)

    @property
    def pos_shftenc(self):
        """
        Read the shftenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('SHFTENC')

    @pos_shftenc.setter
    def pos_shftenc(self, value):
        """
        Set the shftenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('SHFTENC', value)

    @property
    def pos_tgtenc(self):
        """
        Read the tgtenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('TGTENC')

    @pos_tgtenc.setter
    def pos_tgtenc(self, value):
        """
        Set the tgtenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('TGTENC', value)

    @property
    def pos_ctrlenc(self):
        """
        Read the ctrlenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('CTRLENC')

    @pos_ctrlenc.setter
    def pos_ctrlenc(self, value):
        """
        Set the ctrlenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('CTRLENC', value)

    @property
    def pos_encin(self):
        """
        Read the encin register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('ENCIN')

    @pos_encin.setter
    def pos_encin(self, value):
        """
        Set the encin register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('ENCIN', value)

    @property
    def pos_inpos(self):
        """
        Read the inpos register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('INPOS')

    @pos_inpos.setter
    def pos_inpos(self, value):
        """
        Set the inpos register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('INPOS', value)

    @property
    def pos_absenc(self):
        """
        Read the absenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('ABSENC')

    @pos_absenc.setter
    def pos_absenc(self, value):
        """
        Set the absenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('ABSENC', value)

    @property
    def pos_motor(self):
        """
        Read the motor register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('MOTOR')

    @pos_motor.setter
    def pos_motor(self, value):
        """
        Set the motor register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('MOTOR', value)

    @property
    def pos_sync(self):
        """
        Read the sync register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_pos('SYNC')

    @pos_sync.setter
    def pos_sync(self, value):
        """
        Set the sync register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_pos('SYNC', value)

    @property
    def enc(self):
        """
        Read the axis nominal position pointer (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('AXIS')

    @enc.setter
    def enc(self, value):
        """
        Set the axis nominal position value (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('AXIS', value)

    @property
    def enc_shftenc(self):
        """
        Read the shftenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('SHFTENC')

    @enc_shftenc.setter
    def enc_shftenc(self, value):
        """
        Set the shftenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('SHFTENC', value)

    @property
    def enc_tgtenc(self):
        """
        Read the tgtenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('TGTENC')

    @enc_tgtenc.setter
    def enc_tgtenc(self, value):
        """
        Set the tgtenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('TGTENC', value)

    @property
    def enc_ctrlenc(self):
        """
        Read the ctrlenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('CTRLENC')

    @enc_ctrlenc.setter
    def enc_ctrlenc(self, value):
        """
        Set the ctrlenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('CTRLENC', value)

    @property
    def enc_encin(self):
        """
        Read the encin register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('ENCIN')

    @enc_encin.setter
    def enc_encin(self, value):
        """
        Set the encin register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('ENCIN', value)

    @property
    def enc_inpos(self):
        """
        Read the inpos register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('INPOS')

    @enc_inpos.setter
    def enc_inpos(self, value):
        """
        Set the inpos register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('INPOS', value)

    @property
    def enc_absenc(self):
        """
        Read the absenc register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('ABSENC')

    @enc_absenc.setter
    def enc_absenc(self, value):
        """
        Set the absenc register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('ABSENC', value)

    @property
    def enc_motor(self):
        """
        Read the motor register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('MOTOR')

    @enc_motor.setter
    def enc_motor(self, value):
        """
        Set the motor register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('MOTOR', value)

    @property
    def enc_sync(self):
        """
        Read the sync register (IcePAP user manual pag. 108).

        :return: int
        """
        return self.get_enc('SYNC')

    @enc_sync.setter
    def enc_sync(self, value):
        """
        Set the sync register (IcePAP user manual pag. 108).

        :param value: int
        """
        self.set_enc('SYNC', value)

    @property
    def velocity(self):
        """
        Read the nominal velocity (see get_velocity method).

        :return: float
        """
        return self.get_velocity()

    @velocity.setter
    def velocity(self, value):
        """
        Set the nominal velocity (see get_velocity method).

        :param value: float
        """
        self.set_velocity(value)

    @property
    def velocity_min(self):
        """
        Read the minimum velocity (see get_velocity method).

        :return: float
        """
        return self.get_velocity(vtype='MIN')

    @property
    def velocity_max(self):
        """
        Read the maximum velocity (see get_velocity method).

        :return: float
        """
        return self.get_velocity(vtype='MAX')

    @property
    def velocity_default(self):
        """
        Read the default velocity (see get_velocity method).

        :return: float
        """
        return self.get_velocity(vtype='DEFAULT')

    @property
    def velocity_current(self):
        """
        Read the default velocity (see get_velocity method).

        :return: float
        """
        return self.get_velocity(vtype='CURRENT')

    @property
    def acctime(self):
        """
        Get the acceleration time (see get_acceleration method).

        :return: float
        """
        return self.get_acceleration()

    @acctime.setter
    def acctime(self, value):
        """
        Set the acceleration time (see set_acceleration method).

        :param value: float
        """
        self.set_acceleration(value)

    @property
    def acctime_steps(self):
        """
        Get the acceleration in steps distances (see get_acceleration method).

        :return: float
        """
        return self.get_acceleration(atype='STEPS')

    @property
    def acctime_default(self):
        """
        Get the default acceleration time (see get_acceleration method).

        :return: float
        """
        return self.get_acceleration(atype='DEFAULT')

    @property
    def pcloop(self):
        """
        Read if the position close loop is enabled (IcePAP user manual
        pag. 105).

        :return: bool
        """
        return self.send_cmd('?PCLOOP')[0] == 'ON'

    @pcloop.setter
    def pcloop(self, value):
        """
        Activate/Deactivate the position close loop (IcePAP user manual
        pag. 105)

        :param value: bool
        """
        cmd = 'PCLOOP {0}'.format(['OFF', 'ON'][value])
        self.send_cmd(cmd)

    @property
    def jog_velocity(self):
        """
        Get the current jog velocity.

        :return: float
        """
        return float(self.send_cmd('?JOG')[0])

    @property
    def indexer(self):
        """
        Get the indexer signal source used for the axis indexer (IcePAP user
        manual pag. 81).

        :return: str
        """
        return self.send_cmd('?INDEXER')[0]

    @indexer.setter
    def indexer(self, value):
        """
        Set the indexer signal source (IcePAP user manual pag. 81).

        :param value: str
        """
        cmd = 'INDEXER {0}'.format(value)
        self.send_cmd(cmd)

    @property
    def parpos(self):
        """
        Get the position on parametric units (IcePAP user manual pag. 103).

        :return: float
        """
        return float(self.send_cmd('?PARPOS')[0])

    @property
    def parvel(self):
        """
        Get the parametric axis velocity (IcePAP user manual pag. 104).

        :return: float
        """
        return float(self.send_cmd('?PARVEL')[0])

    @parvel.setter
    def parvel(self, value):
        """
        Set the parametric axis velocity (IcePAP user manual pag. 104).

        :param value: float
        """
        # NOTE: SOMETIMES PARVEL 10 RETURNS EXCEPTION:
        # xx:PARVEL ERROR Out of range parameter(s)
        # AND IS AVOIDED BY SETTING IT FIRST TO 0 !!!
        values = [0, value]
        for v in values:
            cmd = 'PARVEL {0}'.format(v)
            self.send_cmd(cmd)

    @property
    def paracct(self):
        """
        Get the parametric acceleration time (IcePAP user manual pag. 99).

        :return: float
        """
        return float(self.send_cmd('?PARACCT')[0])

    @paracct.setter
    def paracct(self, value):
        """
        Set the parametric acceleration time (IcePAP user manual pag. 99).

        :param value: float
        """
        cmd = 'PARACCT {0}'.format(value)
        self.send_cmd(cmd)

    @property
    def homestat(self):
        """
        Return the homing procedure status (IcePAP user manual pag. 79).

        :return: [str(state), int(direction)]
        """
        status, direction = self.send_cmd('?HOMESTAT')
        return status, int(direction)

    @property
    def srchstat(self):
        """
        Return the homing procedure status.
        (IcePAP user manual page 132, v1.0c).

        :return: [str(state), int(direction)]
        """
        status, direction = self.send_cmd('?SRCHSTAT')
        return status, int(direction)

    @property
    def ecam(self):
        """
        Get the electronic cam mode (IcePAP user manual pag. 64).

        :return: str
        """
        return ' '.join(self.send_cmd('?ECAM'))

    @ecam.setter
    def ecam(self, output):
        """
        Set the electronic cam output: OFF, PULSE, LOW, HIGH (IcePAP user
        manual pag. 64).

        :param output: str
        """
        cmd = 'ECAM {0}'.format(output)
        self.send_cmd(cmd)

    @property
    def infoa(self):
        """
        Get the InfoA configuration (Icepap user manual pag. 82).

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?INFOA')

    @infoa.setter
    def infoa(self, cfg):
        """
        Set the InfoA configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'INFOA {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

    @property
    def infob(self):
        """
        Get the InfoB configuration (Icepap user manual pag. 82).

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?INFOB')

    @infob.setter
    def infob(self, cfg):
        """
        Set the InfoB configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'INFOB {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

    @property
    def infoc(self):
        """
        Get the InfoC configuration (Icepap user manual pag. 82).

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?INFOC')

    @infoc.setter
    def infoc(self, cfg):
        """
        Set the InfoC configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'INFOC {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

    @property
    def outpos(self):
        """
        Get the OutPos configuration (Icepap user manual pag. 98)

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?OUTPOS')

    @outpos.setter
    def outpos(self, cfg):
        """
        Set the OutPos configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'OUTPOS {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

    @property
    def outpaux(self):
        """
        Get the OutPAux configuration (Icepap user manual pag. 97).

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?OUTPAUX')

    @outpaux.setter
    def outpaux(self, cfg):
        """
        Set the OutPAux configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'OUTPAUX {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

    @property
    def syncpos(self):
        """
        Get the SyncPos configuration (Icepap user manual pag. 134).

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?SYNCPOS')

    @syncpos.setter
    def syncpos(self, cfg):
        """
        Set the SyncPos configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'SYNCPOS {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

    @property
    def syncaux(self):
        """
        Get the SyncAux configuration (Icepap user manual pag. 82).

        :return: (str, str) [Signal, Polarity]
        """
        return self.send_cmd('?SYNCAUX')

    @syncaux.setter
    def syncaux(self, cfg):
        """
        Set the SyncAux configuration.

        :param cfg: (str, str) [Signal, Polarity]
        """
        cmd = 'SYNCAUX {0} {1}'.format(*cfg)
        self.send_cmd(cmd)

# ------------------------------------------------------------------------
#                       Commands
# ------------------------------------------------------------------------

    def blink(self, secs):
        """
        Blink the driver status led for a given amount of time (Icepap user
        manual pag. 53).

        :param secs: number of second to blink.
        """
        cmd = "BLINK %d" % secs
        self.send_cmd(cmd)

    def send_cmd(self, cmd):
        """
        Wrapper to add the axis number.

        :param cmd: Command without axis number
        :return: [str]
        """
        cmd = '{0}:{1}'.format(self._axis_nr, cmd)
        return self._ctrl.send_cmd(cmd)

    def get_cfginfo(self, parameter=''):
        """
        Get the configuration type for one or all parameters.

        :param parameter: str (optional)
        :return: dict
        """
        cmd = '?CFGINFO {0}'.format(parameter)
        ans = self.send_cmd(cmd)
        cfg = collections.OrderedDict()
        if parameter == '':
            for line in ans:
                key, value = line.split(' ', 1)
                cfg[key] = value
        else:
            key = ans[0]
            value = ' '.join(ans[1:])
            cfg[key] = value
        return cfg

    def set_config(self, config=''):
        """
        Set configuration (IcePAP user manual pag. 59).

        :param config: str
        """
        cmd = 'CONFIG {0}'.format(config)
        self.send_cmd(cmd)

    def get_cfg(self, parameter=''):
        """
        Get the current configuration for one or all parameters (IcePAP user
        manual pag. 54).

        :param parameter: str (optional)
        :return: dict
        """
        cmd = '?CFG {0}'.format(parameter)
        ans = self.send_cmd(cmd)
        cfg = collections.OrderedDict()
        if parameter.lower() in ['', 'default']:
            for line in ans:
                key, value = line.split(' ', 1)
                cfg[key] = value
        else:
            key, value = ans
            cfg[key] = value
        return cfg

    def set_cfg(self, *args):
        """
        Set the configuration of a parameter or change to Default/Expert
        configuration (IcePAP user manual pag. 54).

        set_cfg('Active', 'YES', 'NVOLT', '48',...)

        :param args: str: parameter, value or ('Default')
        """
        cmd = 'CFG {0}'.format(' '.join(args))
        self.send_cmd(cmd)

    def meas(self, parameter):
        """
        Return a measured value for the parameter (IcePAP user manual pag. 89).

        :param parameter: str
        :return: float
        """
        cmd = '?MEAS {0}'.format(parameter)
        return float(self.send_cmd(cmd)[0])

    def get_pos(self, register):
        """
        Read the position register in axis units (IcePAP user manual pag. 108).

        :param register: str
        :return: int
        """
        cmd = '?POS {0}'.format(register)
        return int(self.send_cmd(cmd)[0])

    def set_pos(self, register, value):
        """
        Set the position register in axis units (IcePAP user manual pag. 108).

        :param register: str
        :param value: int
        """
        cmd = 'POS {0} {1}'.format(register, int(value))
        self.send_cmd(cmd)

    def get_enc(self, register):
        """
        Read the position register in encoder step (IcePAP user manual
        pag. 68).

        :param register: str
        :return: int
        """
        cmd = '?ENC {0}'.format(register)
        return int(self.send_cmd(cmd)[0])

    def set_enc(self, register, value):
        """
        Set the position register in encoder step (IcePAP user manual pag. 68).

        :param register: str
        :param value: int
        """
        cmd = 'ENC {0} {1}'.format(register, int(value))
        self.send_cmd(cmd)

    def get_velocity(self, vtype=''):
        """
        Get the velocity (IcePAP user manual pag. 143).

        :param vtype: str
        :return: float [steps per second]
        """
        cmd = '?VELOCITY {0}'.format(vtype)
        return float(self.send_cmd(cmd)[0])

    def set_velocity(self, value):
        """
        Set the velocity (IcePAP user manual pag. 143).

        :param value: float
        :return: float [steps per second]
        """
        cmd = 'VELOCITY {0}'.format(value)
        self.send_cmd(cmd)

    def get_acceleration(self, atype=''):
        """
        Read the acceleration time (IcePAP user manual pag. 48).

        :return: float [seconds]
        """
        cmd = '?ACCTIME {0}'.format(atype)
        return float(self.send_cmd(cmd)[0])

    def set_acceleration(self, value):
        """
        Set the acceleration time (IcePAP user manual pag. 48).

        :param value: float
        """
        cmd = 'ACCTIME {0}'.format(value)
        self.send_cmd(cmd)

    def get_home_position(self, register='AXIS'):
        """
        Return the home value latched on the position register.
        (IcePAP user manual page 83, v1.0c).

        :param register: str
        :return: int
        """
        cmd = '?HOMEPOS {0}'.format(register)
        return int(self.send_cmd(cmd)[0])

    def get_home_encoder(self, register='AXIS'):
        """
        Return the home value latched on encoder register.
        (IcePAP user manual page 82, v1.0c).

        :param register: str
        :return: int
        """
        cmd = '?HOMEENC {0}'.format(register)
        return int(self.send_cmd(cmd)[0])

    def get_srch_position(self, register='AXIS'):
        """
        Return the search value latched on the position register.
        (IcePAP user manual page 131, v1.0c).

        :param register: str
        :return: int
        """
        cmd = '?SRCHPOS {0}'.format(register)
        return int(self.send_cmd(cmd)[0])

    def get_srch_encoder(self, register='AXIS'):
        """
        Return the search value latched on encoder register.
        (IcePAP user manual page 130, v1.0c).

        :param register: str
        :return: int
        """
        cmd = '?SRCHENC {0}'.format(register)
        return int(self.send_cmd(cmd)[0])

    def move(self, position):
        """
        Start absolute movement (IcePAP user manual pag. 92).

        :param position: int
        """
        cmd = 'MOVE {0}'.format(int(position))
        self.send_cmd(cmd)

    def umove(self, position):
        """
        Start absolute updated movement (IcePAP user manual pag. 140).

        :param position: int
        """
        cmd = 'UMOVE {0}'.format(int(position))
        self.send_cmd(cmd)

    def rmove(self, position):
        """
        Start absolute relative movement (IcePAP user manual pag. 140).

        :param position: int
        """
        cmd = 'RMOVE {0}'.format(int(position))
        self.send_cmd(cmd)

    def esync(self):
        """
        Synchronize internal position registers.
        """
        self.send_cmd('ESYNC')

    def ctrlrst(self):
        """
        Reset control encoder value.
        """
        self.send_cmd('CTRLRST')

    def jog(self, veloctiy):
        """
        Start the jog mode at the given velocity (IcePAP user manual pag. 85).

        :param veloctiy: float
        """
        cmd = 'JOG {0}'.format(veloctiy)
        self.send_cmd(cmd)

    def stop(self):
        """
        Stop the current movement with a normal deceleration ramp (IcePAP user
        manual pag. 129).
        """
        self.send_cmd('STOP')

    def abort(self):
        """
        Abort the current movement (IcePAP user manual pag. 46).
        """
        self.send_cmd('ABORT')

    def home(self, mode=1):
        """
        Start home signal search sequence (Icepap user manual pag. 76).

        :param mode: int [-1, 0, 1]
        """
        cmd = 'HOME {0}'.format(mode)
        self.send_cmd(cmd)

    def srch(self, signal, edge_type, direction):
        """
        Start search sequence.
        (IcePAP user manual page 129, v1.0c).

        :param signal: str ['LIM-', 'LIM+', 'HOME', 'ENCAUX', 'INPAUX']
        :param edge_type: str ['POSEDGE', 'NEGEDGE']
        :param direction: int [-1, 1]
        """
        # TODO: Investigate why it does not work with '1' like home command
        cmd = 'SRCH {0} {1} {2:+}'.format(signal, edge_type, direction)
        self.send_cmd(cmd)

    def movel(self, lpos):
        """
        Start postion list movement (IcePAP user manual pag. 93).

        :param lpos: int
        """
        cmd = 'MOVEL {0}'.format(lpos)
        self.send_cmd(cmd)

    def pmove(self, pos):
        """
        Start parametric movement (IcePAP user manual pag. 106).

        :param pos: float
        """
        cmd = 'PMOVE {0}'.format(pos)
        self.send_cmd(cmd)

    def movep(self, pos):
        """
        Start axis movement to a parameter value.

        :param pos: float
        """
        cmd = 'MOVEP {0}'.format(pos)
        self.send_cmd(cmd)

    def cmove(self, pos):
        """
        Start absolute movement in configuration mode (IcePAP user manual
        pag.58).

        :param pos: float
        """
        cmd = 'CMOVE {0}'.format(pos)
        self.send_cmd(cmd)

    def cjog(self, vel):
        """
        Set jog velocity in configuration mode.

        :param vel: float
        """
        cmd = 'CJOG {0}'.format(vel)
        self.send_cmd(cmd)

    def track(self, signal, mode='FULL'):
        """
        Start position tracking mode (IcePAP user manual pag. 139).

        :param signal: str
        :param mode: str
        """
        cmd = 'TRACK {0} {1}'.format(signal, mode)
        self.send_cmd(cmd)

    def ptrack(self, signal, mode='FULL'):
        """
        Start parametric tracking mode (IcePAP user manual pag. 114).

        :param signal: str
        :param mode: str
        """
        cmd = 'PTRACK {0} {1}'.format(signal, mode)
        self.send_cmd(cmd)

    def set_ecam_table(self, lpos, source='AXIS', dtype='FLOAT'):
        """
        Load the position list to the electronic cam table. The maximum memory
        size of the table is 81908 bytes. After to load the electronic cam
        is ON (IcePAP user manual pag. 65).

        :param lpos: [float]
        :param source: str
        :param dtype: str
        """
        # Sort the given list of points in ascending order
        # (<list>.sort() slightly more efficient:
        lpos.sort()
        lushorts = self.get_ushort_list(lpos, dtype)
        if len(lushorts) > 40954:
            raise ValueError('There is not enough memory to load the list.')

        cmd = '*ECAMDAT {0} {1}'.format(source, dtype)
        self.send_cmd(cmd)
        self._ctrl._comm.send_binary(lushorts)

    def clear_ecam_table(self):
        """
        Clean the electronic cam table (Icepap user manual pag. 65).
        """
        self.send_cmd('ECAMDAT CLEAR')

    def get_ecam_table(self, dtype='FLOAT'):
        """
        Get the position table loaded on the electronic cam table (IcePAP user
        manual pag. 65).

        :param dtype: str
        :return: [float]
        """
        cmd = '?ECAMDAT {0} {1}'
        return self._get_dump_table(cmd, dtype)

    def set_list_table(self, lpos, cyclic=False, dtype='FLOAT'):
        """
        Load a position list table (Icepap user manual pag. 87).

        :param lpos: [float]
        :param cyclic: bool
        :param dtype: str
        """
        lushorts = self.get_ushort_list(lpos, dtype)
        # if len(lushorts) > 40954:
        #     raise ValueError('There is not enough memory to load the list.')

        cmd = '*LISTDAT {0} {1}'.format(['NOCYCLIC', 'CYCLIC'][cyclic], dtype)
        self.send_cmd(cmd)
        self._ctrl._comm.send_binary(lushorts)

    def clear_list_table(self):
        """
        Clean the position list table (Icepap user manual pag. 87).
        """
        self.send_cmd('LISTDAT CLEAR')

    def get_list_table(self, dtype='FLOAT'):
        """
        Get the position list table (IcePAP user manual pag. 87).

        :param dtype: str
        :return: [float]
        """
        cmd = '?LISTDAT {0} {1}'
        return self._get_dump_table(cmd, dtype)

    def set_parametric_table(self, lparam, lpos, lslope=None, mode='SPLINE',
                             param_type=FLOAT, pos_type=DWORD,
                             slope_type=FLOAT):
        """
        Method to set the parametric trajectory data (IcePAP user manual
        pag. 100).

        :param lparam: [float]
        :param lpos: [int]
        :param lslope: [float]
        :param mode: str [Linear, Spline, Cyclic]'
        :param param_type: str (Global Definitions in icepap.vdatalib)
        :param pos_type: str (Global Definitions in icepap.vdatalib)
        :param slope_type: str (Global Definitions in icepap.vdatalib)
        """
        data = vdata()
        data.append(lparam, ADDRUNSET, PARAMETER, format=param_type)
        data.append(lpos, self._axis_nr, POSITION, format=pos_type)
        if lslope is not None:
            data.append(lslope, self.addr, SLOPE, format=slope_type)

        bin_data = data.bin().flatten()
        lushorts = self.get_ushort_list(bin_data, dtype='BYTE')
        cmd = '*PARDAT {0}'.format(mode)
        self.send_cmd(cmd)
        self._ctrl._comm.send_binary(lushorts)

    def clear_parametric_table(self):
        """
        Clean the parametric trajectory table.
        """
        self.send_cmd('PARDAT CLEAR')

    def get_parametric_table(self):
        """
        Get the parametric table.

        :return: [[float]] ([parameter], [position], [slope])
        """
        MAX_SUBSET_SIZE = 20
        nr_points = int(self.send_cmd('?PARDAT NPTS')[0])
        if nr_points < MAX_SUBSET_SIZE:
            MAX_SUBSET_SIZE = nr_points
        if MAX_SUBSET_SIZE == 0:
            raise RuntimeError('There are not vlaues loaded on the '
                               'parametric table')
        lparam = []
        lpos = []
        lslope = []
        start_pos = 0
        cmd = '?PARDAT {0} {1}'
        packages = nr_points // MAX_SUBSET_SIZE
        for i in range(packages):
            raw_values = self.send_cmd(cmd.format(start_pos, MAX_SUBSET_SIZE))
            start_pos += MAX_SUBSET_SIZE
            for raw_value in raw_values:
                param, pos, slope = raw_value.strip().split()
                lparam.append(float(param))
                lpos.append(float(pos))
                lslope.append(float(slope))
        if (nr_points % MAX_SUBSET_SIZE) > 0:
            last_points = nr_points - start_pos
            raw_values = self.send_cmd(cmd.format(start_pos, last_points))
            for raw_value in raw_values:
                param, pos, slope = raw_value.strip().split()
                lparam.append(float(param))
                lpos.append(float(pos))
                lslope.append(float(slope))

        return lparam, lpos, lslope

    def get_parval(self, parameter):
        """
        Get the motor position for a parameter value.

        :param parameter: float
        :return: float
        """
        cmd = '?PARVAL {0}'.format(parameter)
        return float(self.send_cmd(cmd)[0])

    def print_commands(self):
        """
        Get the allows commands (IcePAP user manual pag. 75).
        """
        ans = self.send_cmd('?HELP')
        print(('\n'.join(ans)))



# !/usr/bin/env python
# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import argparse
import logging
import logging.config
import time
import os
import sys
from .backups import IcePAPBackup
from .communication import IcePAPCommunication
from .programming import firmware_update
from .__init__ import version


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        },
        'simple': {
            'format': '%(levelname)-8s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'file': {
            'level': 'INFO',
            'class': 'logging.FileHandler',
            'filename': '',
            'mode': 'w',
            'encoding': 'utf-8',
            'formatter': 'verbose'
        }
    },
    'loggers': {
        'Application': {
            'handlers': ['console', 'file'],
            'propagate': True,
            'level': 'INFO',
        },
        'icepap': {
            'handlers': ['file'],
            'level': 'INFO',
            'propagate': True,
        }
    }
}


def end(log, err_no=0):
    log_file = LOGGING_CONFIG['handlers']['file']['filename']
    log.info('Log saved on: {0}'.format(log_file))
    for h in log.handlers:
        h.flush()
    sys.exit(err_no)


def get_parser():
    desc = 'IcePAP scripts, base on ethernet communication\n'
    desc += 'Version: {}.\n'.format(version)
    epi = 'Documentation: https://alba-synchrotron.github.io/pyIcePAP-doc/\n'
    epi += 'Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain.'
    fmt = argparse.RawTextHelpFormatter
    parse = argparse.ArgumentParser(description=desc,
                                    formatter_class=fmt,
                                    epilog=epi)
    ver = '%(prog)s {0}'.format(version)
    parse.add_argument('--version', action='version', version=ver)
    subps = parse.add_subparsers(help='commands')

    # -------------------------------------------------------------------------
    #                           Backup commands
    # -------------------------------------------------------------------------
    # Save backup command
    save_cmd = subps.add_parser('save', help='Save the '
                                             'configuration to a file')
    save_cmd.set_defaults(which='save')
    save_cmd.add_argument('host', help='IcePAP Host')
    save_cmd.add_argument('-p', '--port', default=5000, help='IcePAP port')
    save_cmd.add_argument('-t', '--timeout', default=3, help='Socket timeout')
    save_cmd.add_argument('--bkpfile', help='Output backup filename',
                          default='')
    save_cmd.add_argument('axes', nargs='*', help='Axes to save, default all',
                          type=int, default=[])
    save_cmd.add_argument('-d', '--debug', action='store_true',
                          help='Activate log level DEBUG')

    # Check backup command
    check_cmd = subps.add_parser('check', help='Check the '
                                               'IcePAP configuration for a '
                                               'backup file')
    check_cmd.set_defaults(which='check')
    check_cmd.add_argument('filename', help='Backup file')
    check_cmd.add_argument('axes', nargs='*', help='Axes to save, default all',
                           type=int, default=[])
    check_cmd.add_argument('-d', '--debug', action='store_true',
                           help='Activate log level DEBUG')
    check_cmd.add_argument('--host',
                           help='Use another IcePAP host instead of the '
                                'backup saved.',
                           default='')

    # -------------------------------------------------------------------------
    #                           Firmware commands
    # -------------------------------------------------------------------------
    # Update
    update_cmd = subps.add_parser('update', help='Change the '
                                                 'firmware version. It '
                                                 'creates a backup before to '
                                                 'change the FW')
    update_cmd.set_defaults(which='update')
    update_cmd.add_argument('host', help='IcePAP Host')
    update_cmd.add_argument('-p', '--port', default=5000, help='IcePAP port')
    update_cmd.add_argument('-t', '--timeout', default=3,
                            help='Socket timeout')
    update_cmd.add_argument('--bkpfile', help='Output backup filename',
                            default='')
    update_cmd.add_argument('--no-check',
                            help='Avoid the checking procedure after the '
                                 'update',
                            dest='nocheck')
    update_cmd.add_argument('fwfile', help='Firmware binary file')
    update_cmd.add_argument('-d', '--debug', action='store_true',
                            help='Activate log level DEBUG')

    # Autofix
    autofix_cmd = subps.add_parser('autofix', help='Autofix a driver '
                                                   'configuration')
    autofix_cmd.set_defaults(which='autofix')
    autofix_cmd.add_argument('host', help='IcePAP Host')
    autofix_cmd.add_argument('-p', '--port', default=5000, help='IcePAP port')
    autofix_cmd.add_argument('-t', '--timeout', default=3,
                             help='Socket timeout')
    autofix_cmd.add_argument('--bkpfile', help='Output backup filename',
                             default='')
    autofix_cmd.add_argument('--no-check',
                             help='Avoid the checking procedure after the '
                                  'update',
                             dest='nocheck')
    autofix_cmd.add_argument('loadfile', help='backup to load')
    autofix_cmd.add_argument('-d', '--debug', action='store_true',
                             help='Activate log level DEBUG')

    # -------------------------------------------------------------------------
    #                           IcePAP commands
    # -------------------------------------------------------------------------
    # Send raw command
    send_cmd = subps.add_parser('send', help='Send IcePAP raw '
                                             'commands')

    send_cmd.set_defaults(which='send')
    send_cmd.add_argument('host', help='IcePAP Host')
    send_cmd.add_argument('-p', '--port', default=5000, help='IcePAP port')
    send_cmd.add_argument('-t', '--timeout', default=3, help='Socket timeout')
    send_cmd.add_argument('command', help='Raw command',
                          type=str)
    send_cmd.add_argument('-d', '--debug', action='store_true',
                          help='Activate log level DEBUG')
    # -------------------------------------------------------------------------
    return parse


def get_filename(host, command, filename='', log=False):
    value = time.strftime('%Y%m%d_%H%m%S')
    ext = ['cfg', 'log'][log]
    if host == '':
        host = 'icepap'
    new_filename = '{0}_{1}_{2}.{3}'.format(value, host, command, ext)
    if log or filename == '':
        filename = new_filename
    return os.path.abspath(filename)


def main():
    args = get_parser().parse_args()

    log_file = get_filename(args.host, args.which, log=True)
    LOGGING_CONFIG['handlers']['file']['filename'] = log_file
    logging.config.dictConfig(LOGGING_CONFIG)
    log = logging.getLogger('Application')

    # Save Command
    if args.which == 'save':
        abspath = get_filename(args.host, args.which, args.bkpfile)
        log.info('Saving backup on: {0}'.format(abspath))
        ipap_bkp = IcePAPBackup(args.host, args.port, args.timeout)
        ipap_bkp.do_backup(abspath, args.axes)

    # Check Command
    elif args.which == 'check':
        ipap_bkp = IcePAPBackup(host=args.host, cfg_file=args.filename)
        ipap_bkp.do_check(args.axes)

    # Send Command
    elif args.which == 'send':
        ipap_com = IcePAPCommunication(host=args.host, port=args.port,
                                       timeout=args.timeout)
        log.info(ipap_com.send_cmd(args.command))

    # Update Command
    elif args.which == 'update':
        abspath = get_filename(args.host, args.which, args.bkpfile)
        log.info('Updating {0} with firmware file {1}'.format(args.host,
                                                              args.fwfile))
        log.info('Saving backup on: {0}'.format(abspath))
        ipap_bkp = IcePAPBackup(host=args.host, port=args.port,
                                timeout=args.timeout)
        ipap_bkp.do_backup(abspath)
        if not firmware_update(args.host, args.fwfile, log):
            log.error('Errors on update firmware. Skipping active and check '
                      'driver procedure.')
            end(log, -1)

        log.info('Restore active drivers...')
        ipap_bkp = IcePAPBackup(cfg_file=abspath)
        ipap_bkp.active_axes()
        if args.nocheck:
            end(log)

        log.info('Checking registers....')
        diff = ipap_bkp.do_check()
        if len(diff) == 0:
            log.info('No differences found')
            end(log)

        log.info('Fixing differences')
        ipap_bkp.do_autofix(diff)
        end(log)

    elif args.which == 'autofix':
        abspath = get_filename(args.host, args.which, args.bkpfile)
        log.info('Saving backup on: {0}'.format(abspath))
        ipap_bkp = IcePAPBackup(host=args.host, port=args.port,
                                timeout=args.timeout)
        ipap_bkp.do_backup(abspath)

        ipap_bkp = IcePAPBackup(host=args.host, cfg_file=args.loadfile)
        log.info('Checking registers....')
        diff = ipap_bkp.do_check()
        if len(diff) == 0:
            log.info('No differences found')
            end(log)

        log.info('Fixing differences')
        ipap_bkp.do_autofix(args.axes)
        end(log)


if __name__ == '__main__':
    main()



# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------
import sys
import time
import array
from icepap import IcePAPController

__all__ = ['firmware_update']

# TODO: Review downgrade from 1.225 to 1.22 (MCPU<X> components not updated)


def _monitor(ice):
    wait = True
    try:
        if not ice.connected:
            raise RuntimeError('No connection available')
        status = ice.get_prog_status()
        if status[0].upper() != 'DONE':
            _progress_bar(float(status[-1]), 100, 'Updating IcePAP firmware')
            time.sleep(.2)
        else:
            wait = False
    except RuntimeError:
        print('\nConnection lost, waiting for reconnection...')
        while not ice.connected:
            time.sleep(0.5)
        print('Reconnected!')
    finally:
        return wait


def load_firmware(ice, filename):
    """
    Loads code firmware contained in filename to system master controller
    internal flash memory (tocho file).

    :param ice: IcePAPController object.
    :param filename: firmware code.
    """
    if filename:
        with open(filename, 'rb') as f:
            data = f.read()
        data = array.array('H', data)
        ice.sprog(saving=True)
        ice._comm.send_binary(ushort_data=data)

    # Avoid timeout while ICEPAP is saving the file to master
    time.sleep(5)
    # print('New firmware loaded to master')


def _progress_bar(count, total, status=''):
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s < %s >\r' % (bar, percents, '%', status))
    sys.stdout.flush()
    # As suggested by Rom Ruben (see: http://stackoverflow.com/questions/
    # 3173320/text-progress-bar-in-the-console/
    # 27871113#comment50529068_27871113)


def install_firmware(ice, component='ALL', force=False, saving=False,
                     filename=None):
    """
    Update icepap firmware. If filename is not supplied, the firmware
    update will use the code stored in the system master controller.

    :param ice: IcePAPController object.
    :param component: { board adress | DRIVERS | CONTROLLERS| ALL }
    :param force: Force overwrite regardless of being idential versions.
    :param saving: Saves firmware into master board flash.
    :param filename: firmware code.
    """

    if filename:
        with open(filename, 'rb') as f:
            data = f.read()
        data = array.array('H', data)
        # from external firmware code
        ice.sprog(component, force=force, saving=saving)
        # TODO: make the method public
        ice._comm.send_binary(ushort_data=data)
    else:
        ice.prog(component, force=force)


# TODO: Define entry point for system_update method

def firmware_update(hostname, filename, log):
    """
    Installs firmware stored in master memory to ALL system components.

    :param hostname: Icepap host
    :param filename: Firmware filename
    :param log: Logger object
    """
    ice = IcePAPController(hostname)

    try:
        curr_ver = ice.ver['SYSTEM']['VER'][0]
    except Exception as e:
        log.error('Can not read the current version. {}'.format(e))
        curr_ver = -1
    log.info('Current firmware version is {}'.format(curr_ver))
    ice.mode = 'prog'
    load_firmware(ice, filename)
    log.info('Firmware "{}" loaded to master.'.format(filename))
    try:
        new_ver = ice.ver_saved.system[0]
    except Exception:
        new_ver = -1

    log.info('Installing version {}'.format(new_ver))
    if curr_ver < 3.17:

        install_firmware(ice, 'ALL', force=True)
        install_firmware(ice, 'MCPU0')
        install_firmware(ice, 'MCPU1')
        install_firmware(ice, 'MCPU2')
    else:
        install_firmware(ice, 'ALL', force=True)

    # wait process to finish
    wait = True
    while wait:
        wait = _monitor(ice)
        time.sleep(0.5)
    ice.reboot()
    print('')
    # wait reboot to finish
    for i in range(120):
        _progress_bar(i+1, 120, 'Rebooting IcePAP')
        time.sleep(1)
    print('')
    ice = IcePAPController(hostname)
    ice.mode = 'oper'
    if ice.mode.lower() != 'oper':
        log.error('[ERROR: It was not possible to set mode oper!!!]')
        return False
    log.info('Change firmware DONE')
    return True
